using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using PursuitDrift.Interfaces;
using PursuitDrift.Physics;
using StillDesign.PhysX;

namespace PursuitDrift.GameObjects.PhysicObjects.Cars
{
    public class CarObject : PhysicObject
    {
        private Model _carmodel;
        private Model _wheelmodel;

        private Actor _chassisActor;

        private CarInformation _carInfo;

        private WheelShape _leftFrontWheel;
        private WheelShape _rightFrontWheel;
        private WheelShape _leftRearWheel;
        private WheelShape _rightRearWheel;

        //private const float WheelRadius = 0.4f;
        //private const float ChassisLength = 6.0f;
        //private const float ChassisWidth = 2.3f;
        //private const float ChassisHeight = 1.6f;

        public CarObject(Game game, IPursuitDriftGame baseGame, CarInformation carInfo)
            : base(game, baseGame)
        {
            _carInfo = carInfo;
        }

        public void Reset()
        {
            _chassisActor.GlobalPose = Matrix.Identity * Matrix.CreateTranslation(0, 1, 0);
            _chassisActor.AngularDamping = 0;
            _chassisActor.AngularVelocity = Vector3.Zero;
            _chassisActor.LinearVelocity = Vector3.Zero;

            _leftFrontWheel.MotorTorque = 0;
            _leftFrontWheel.SteeringAngle = 0;
            _leftRearWheel.MotorTorque = 0;
            _leftRearWheel.SteeringAngle = 0;
            _rightFrontWheel.MotorTorque = 0;
            _rightFrontWheel.SteeringAngle = 0;
            _rightRearWheel.MotorTorque = 0;
            _rightRearWheel.SteeringAngle = 0;
        }

        public void Accelerate(float addspeed)
        {
            if (_carInfo.FrontWheelDrive)
            {
                _leftFrontWheel.MotorTorque -= addspeed;
                _rightFrontWheel.MotorTorque -= addspeed;
            }
            if (_carInfo.RearWheelDrive)
            {
                _rightRearWheel.MotorTorque -= addspeed;
                _leftRearWheel.MotorTorque -= addspeed;
            }

            _chassisActor.WakeUp();
        }

        public void Turn(float newangle)
        {
            _leftFrontWheel.SteeringAngle = _rightFrontWheel.SteeringAngle = newangle;
        }

        protected override void LoadContent()
        {
            _carmodel = Game.Content.Load<Model>(_carInfo.CarModelAsset);
            _wheelmodel = Game.Content.Load<Model>(_carInfo.WheelModelAsset);

            InitCarPhysics();

            base.LoadContent();
        }

        private void InitCarPhysics()
        {
            CreateChassis();

            const float yoffset = -0.7f;
            const float frontxoffset = 1.22f;
            const float rearxoffset = 1.2f;
            const float rightzoffset = -0.1f;
            const float leftzoffset = 0.1f;

            Matrix _wheelRotation = Matrix.CreateRotationY(-MathHelper.PiOver2);

            // Rder erstellen
            CreateWheel(out _leftFrontWheel, _wheelRotation * Matrix.CreateTranslation(_carInfo.CarLenght / 2 - frontxoffset, yoffset, _carInfo.CarWidth / 2 - leftzoffset));
            CreateWheel(out _rightFrontWheel, _wheelRotation * Matrix.CreateTranslation(_carInfo.CarLenght / 2 - frontxoffset, yoffset, -(_carInfo.CarWidth / 2) - rightzoffset));
            CreateWheel(out _leftRearWheel, _wheelRotation * Matrix.CreateTranslation(-(_carInfo.CarLenght / 2) + rearxoffset, yoffset, _carInfo.CarWidth / 2 - leftzoffset));
            CreateWheel(out _rightRearWheel, _wheelRotation * Matrix.CreateTranslation(-(_carInfo.CarLenght / 2) + rearxoffset, yoffset, -(_carInfo.CarWidth / 2) - rightzoffset));
        }

        private void CreateChassis()
        {
            BodyDescription _chassisBodyDesc = new BodyDescription(_carInfo.CarMass);
            BoxShapeDescription _chassisShapeDesc = new BoxShapeDescription(_carInfo.CarLenght, _carInfo.CarHeight, _carInfo.CarWidth);

            ActorDescription _chassisDesc = new ActorDescription
                                                {
                                                    BodyDescription = _chassisBodyDesc,
                                                    Shapes = { _chassisShapeDesc },
                                                    GlobalPose = Matrix.CreateTranslation(0, 1, 0)
                                                };

            _chassisActor = PhysX.Scene.CreateActor(_chassisDesc);

            _chassisActor.SetCenterOfMassOffsetLocalPosition(_carInfo.CenterOfMass);
        }

        private void CreateWheel(out WheelShape wheelShape, Matrix localPose)
        {
            BodyDescription _bodyDesc = new BodyDescription(5.0f);

            WheelShapeDescription _wheelDesc = new WheelShapeDescription
                                                   {
                                                       Radius = _carInfo.WheelRadius,
                                                       SuspensionTravel = _carInfo.SuspensionTravel,
                                                       LocalPose = localPose,
                                                   };

            // Not needed
            //ActorDescription _actorDesc = new ActorDescription
            //{
            //    BodyDescription = _bodyDesc,
            //    Shapes = { _wheelDesc }
            //};

            wheelShape = _chassisActor.CreateShape<WheelShape>(_wheelDesc);
        }

        public override void Draw(GameTime gameTime)
        {
            var _cartransforms = new Matrix[_carmodel.Bones.Count];
            _carmodel.CopyAbsoluteBoneTransformsTo(_cartransforms);

            foreach (var _mesh in _carmodel.Meshes)
            {
                foreach (BasicEffect _effect in _mesh.Effects)
                {
                    _effect.PreferPerPixelLighting = true;
                    _effect.EnableDefaultLighting();

                    _effect.View = BaseGame.Camera.View;
                    _effect.Projection = BaseGame.Camera.Projection;
                    _effect.World = _cartransforms[_mesh.ParentBone.Index] *
                                    Matrix.CreateTranslation(0.0f, -(_carInfo.CarHeight / 2), 0.0f) *
                                    _chassisActor.GlobalOrientation *
                                    Matrix.CreateTranslation(_chassisActor.GlobalPosition);
                }

                _mesh.Draw();
            }

            DrawWheel(_leftFrontWheel, WheelPosition.FrontLeft);
            DrawWheel(_leftRearWheel, WheelPosition.RearLeft);
            DrawWheel(_rightFrontWheel, WheelPosition.FrontRight);
            DrawWheel(_rightRearWheel, WheelPosition.RearRight);

            base.Draw(gameTime);
        }

        public override void Update(GameTime gameTime)
        {
            //_leftFrontWheel.SteeringAngle += 0.01f;
            //_leftRearWheel.SteeringAngle += 0.02f;
            //_rightFrontWheel.SteeringAngle += 0.03f;
            //_rightRearWheel.SteeringAngle += 0.04f;

            Game.Window.Title = string.Format("LF-Suspension: {0}", _leftFrontWheel.Suspension.TargetValue);

            base.Update(gameTime);
        }

        private void DrawWheel(WheelShape wheelShape, WheelPosition position)
        {
            bool _right = position == WheelPosition.FrontRight || position == WheelPosition.RearRight;
            float _yrot = _right ? -MathHelper.Pi : 0.0f;

            foreach (ModelMesh _mesh in _wheelmodel.Meshes)
            {
                foreach (BasicEffect _effect in _mesh.Effects)
                {
                    _effect.EnableDefaultLighting();

                    _effect.View = BaseGame.Camera.View;
                    _effect.Projection = BaseGame.Camera.Projection;
                    _effect.World = Matrix.CreateRotationY(_yrot) *
                                    Matrix.CreateRotationZ(wheelShape.AxleSpeed) *
                                    Matrix.CreateTranslation(0.0f, -wheelShape.Suspension.TargetValue, 0.0f) *
                                    Matrix.CreateRotationY(wheelShape.SteeringAngle) *
                                    _chassisActor.GlobalOrientation *
                                    Matrix.CreateTranslation(wheelShape.GlobalPosition) *
                                    Matrix.Identity;

                }
                _mesh.Draw();
            }
        }
    }
}