using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using PursuitDrift.Interfaces;
using StillDesign.PhysX;

namespace PursuitDrift.Physics
{
    /// <summary>
    /// Statische Klasse fr die PhysX.
    /// </summary>
    public static class PhysX
    {
        #region Private Fields

        private static Core _core;
        private static Scene _scene;

        private static BasicEffect _visualizationEffect;

        #endregion

        #region Properties

        /// <summary>
        /// Gibt den PhysX-Core zurck oder legt diesen fest.
        /// </summary>
        public static Core Core
        {
            get { return _core; }
            set { _core = value; }
        }

        /// <summary>
        /// Gibt die PhysX-Scene zurck oder legt diese fest.
        /// </summary>
        public static Scene Scene
        {
            get { return _scene; }
            set { _scene = value; }
        }

        /// <summary>
        /// Gibt einen Wert zurck, der angibt, ob DebugView 
        /// eingeschaltet ist, oder legt diesen fest.
        /// </summary>
        public static bool IsDebugViewEnabled
        {
            get
            {
                bool _vCollShapes = Core.Parameters.VisualizeCollisionShapes;
                bool _vClothMesh = Core.Parameters.VisualizeClothMesh;
                bool _vJointLAxes = Core.Parameters.VisualizeJointLocalAxes;
                bool _vJointLimits = Core.Parameters.VisualizeJointLimits;
                bool _vFluidPos = Core.Parameters.VisualizeFluidPosition;
                bool _vForceFields = Core.Parameters.VisualizeForceFields;
                bool _vSoftBodyMesh = Core.Parameters.VisualizeSoftBodyMesh;
                bool _vWorldAxes = Core.Parameters.VisualizeWorldAxes;

                //bool _remoteDebugger = Core.Foundation.RemoteDebugger.IsConnected;

                return _vCollShapes && _vClothMesh && _vJointLAxes &&
                       _vJointLimits && _vFluidPos && _vForceFields &&
                       _vSoftBodyMesh && _vWorldAxes;
            }
            set
            {
                Core.Parameters.VisualizationScale = 2.0f;

                Core.Parameters.VisualizeCollisionShapes = value;
                Core.Parameters.VisualizeClothMesh = value;
                Core.Parameters.VisualizeJointLocalAxes = value;
                Core.Parameters.VisualizeJointLimits = value;
                Core.Parameters.VisualizeFluidPosition = value;
                Core.Parameters.VisualizeForceFields = value;
                Core.Parameters.VisualizeSoftBodyMesh = value;
                Core.Parameters.VisualizeWorldAxes = value;

                if (value)
                    Core.Foundation.RemoteDebugger.Connect("localhost");
                else
                    Core.Foundation.RemoteDebugger.Disconnect();
            }
        }

        #endregion

        #region Methods

        public static void InitializePhysX(Game game)
        {
            var _coreDesc = new CoreDescription();
            _core = new Core(_coreDesc, null);

            var _sceneDesc = new SceneDescription
                                 {
                                     Gravity = new Vector3(0.0f, -9.81f, 0.0f),
                                     // -9.81 ist die Anziehungskraft auf der Erde
                                     GroundPlaneEnabled = true
                                 };

            _scene = _core.CreateScene(_sceneDesc);

            _visualizationEffect = new BasicEffect(game.GraphicsDevice, null)
                                       {
                                           VertexColorEnabled = true
                                       };
        }

        public static void ToggleDebugView()
        {
            IsDebugViewEnabled = !IsDebugViewEnabled;
        }

        public static void DrawVisualiation(Game game, IPursuitDriftGame baseGame)
        {
            game.GraphicsDevice.VertexDeclaration = new VertexDeclaration(game.GraphicsDevice,
                                                                              VertexPositionColor.VertexElements);

            _visualizationEffect.World = Matrix.Identity;
            _visualizationEffect.View = baseGame.Camera.View;
            _visualizationEffect.Projection = baseGame.Camera.Projection;

            DebugRenderable _data = Scene.GetDebugRenderable();

            _visualizationEffect.Begin();

            foreach (EffectPass pass in _visualizationEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                if (_data.PointCount > 0)
                {
                    DebugPoint[] _points = _data.GetDebugPoints();

                    game.GraphicsDevice.DrawUserPrimitives(PrimitiveType.PointList, _points, 0, _points.Length);
                }

                if (_data.LineCount > 0)
                {
                    DebugLine[] _lines = _data.GetDebugLines();

                    var _vertices = new VertexPositionColor[_data.LineCount * 2];
                    for (int x = 0; x < _data.LineCount; x++)
                    {
                        DebugLine _line = _lines[x];

                        _vertices[x * 2 + 0] = new VertexPositionColor(_line.Point0, Int32ToColor(_line.Color));
                        _vertices[x * 2 + 1] = new VertexPositionColor(_line.Point1, Int32ToColor(_line.Color));
                    }

                    game.GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, _vertices, 0, _lines.Length);
                }

                if (_data.TriangleCount > 0)
                {
                    DebugTriangle[] _triangles = _data.GetDebugTriangles();

                    var _vertices = new VertexPositionColor[_data.TriangleCount * 3];
                    for (int x = 0; x < _data.TriangleCount; x++)
                    {
                        DebugTriangle _triangle = _triangles[x];

                        _vertices[x * 3 + 0] = new VertexPositionColor(_triangle.Point0, Int32ToColor(_triangle.Color));
                        _vertices[x * 3 + 1] = new VertexPositionColor(_triangle.Point1, Int32ToColor(_triangle.Color));
                        _vertices[x * 3 + 2] = new VertexPositionColor(_triangle.Point2, Int32ToColor(_triangle.Color));
                    }

                    game.GraphicsDevice.DrawUserPrimitives(PrimitiveType.TriangleList, _vertices, 0, _triangles.Length);
                }

                pass.End();
            }

            _visualizationEffect.End();
        }

        private static Color Int32ToColor(int color)
        {
            var a = (byte)((color & 0xFF000000) >> 32);
            var r = (byte)((color & 0x00FF0000) >> 16);
            var g = (byte)((color & 0x0000FF00) >> 8);
            var b = (byte)((color & 0x000000FF) >> 0);

            return new Color(r, g, b, a);
        }

        public static void UpdatePhysX(GameTime gameTime)
        {
            _scene.Simulate(gameTime.ElapsedGameTime.Milliseconds / 1000.0f);
            _scene.FlushStream();
            _scene.FetchResults(SimulationStatus.RigidBodyFinished, true);
        }

        #endregion
    }
}